from textual.app import App, ComposeResult
from textual.containers import Container, Horizontal, Vertical
from textual.widgets import Button, Footer, Header, Static, Label, Placeholder, Tree
from textual.messages import Message
from textual.widget import Widget

import girok.api.category as category_api
import girok.utils.calendar as calendar_utils
from rich.table import Table
from rich.style import Style
from textual import log

from girok.calendar_cli.sidebar import SidebarContainer, CategoryTree
from girok.calendar_cli.calendar_container import CalendarContainer
import girok.constants as constants
        
class CalendarApp(Horizontal):
    CSS_PATH = "./demo_dock.css"
    
    def compose(self):
        yield SidebarContainer(id="sidebar-container")
        yield CalendarContainer(id="calendar-container")
                
    def on_category_tree_category_changed(self, event):
        self.query_one(CalendarContainer).update_cat_path(event.cat_path)
        cat_tree = self.query_one(CategoryTree)
        
    def on_tag_tree_tag_changed(self, event):
        tag = event.tag
        if tag.endswith(" " + constants.LEFT_ARROW_EMOJI):
            tag = tag[:-2]
        if tag == "All Tags":
            tag = ""
        self.query_one(CalendarContainer).update_tag(tag)
        
    def on_category_tree_custom_test_message(self, event):
        self.refresh()


DISPLAY_TERMINAL_COLOR_TITLE_TEXT = "#000000"
DISPLAY_TERMINAL_COLOR_TITLE_BACKGROUND = "#F6DA75"
DISPLAY_TERMINAL_COLOR_SUCCESS_TEXT = "#000000"
DISPLAY_TERMINAL_COLOR_SUCCESS_BACKGROUND = "#B0D78F"
DISPLAY_TERMINAL_COLOR_ERROR_TEXT = "#000000"
DISPLAY_TERMINAL_COLOR_ERROR_BACKGROUND = "#FFAFB0"
DISPLAY_TERMINAL_COLOR_WARNING_TEXT = "#000000"
DISPLAY_TERMINAL_COLOR_WARNING_BACKGROUND = "#F2CFA5"

DISPLAY_TERMINAL_COLOR_TITLE = "cyan on purple3"
DISPLAY_TERMINAL_COLOR_SUCCESS = "black on green"
DISPLAY_TERMINAL_COLOR_ERROR = "black on red"
DISPLAY_TERMINAL_COLOR_WARNING = "black on bright_red"

DEFAULT_TASK_COLOR = "grey"

BORDER_COLOR = ""
DEFAULT_BACKGROUND_COLOR = "#171921"
DEFAULT_FONT_COLOR = "#8f929b"
TASK_COLOR = ""
CALENDAR_DATE_COLOR = "red"
CATEGORY_COLOR = ""

# Calendar date
CALENDAR_HEADER_DATE_COLOR = "#F6E2B2"
CALENDAR_WEEKDAY_NAME_COLOR = "#E6E0DB"
CALENDAR_TODAY_COLOR = "#FADD70"


# Pop up box
TABLE_TASK_NAME_COLOR = "#FAC7C1"
TABLE_TEXT_COLOR = "#D7C8B7"
TABLE_HEADER_COLOR = "#D7E1C9"
TABLE_HEADER_DATE_COLOR = "#E8E5DA"
TABLE_DEADLINE_COLOR = "#A8CCDC"
TABLE_TASK_HIGHLIGHT_COLOR = "#B1EEC9"
TABLE_TASK_DELETED_COLOR = "#808080"

# EMOJI
LEFT_ARROW_EMOJI = "⬅"
CIRCLE_EMOJI = "●"


CIRCLE_COLOR = {
    "yellow": "#f9e316",
    "green": "#29f429",
    "orange": "#EF5D0C",
    "white": "#E4DCCE",
    "brown": "#8F4802",
    "blue": "#6893F8",
    "grey": "#808080"
}




import requests
from rich.console import Console

from girok.config import get_config
import girok.utils.auth as auth_utils
import girok.utils.general as general_utils
import girok.utils.display as display_utils
import girok.constants as constants

console = Console()
cfg = get_config()

def get_categories():
    resp = requests.get(
        cfg.base_url + "/categories",
        headers=auth_utils.build_jwt_header(cfg.config_path)
    )
    if resp.status_code == 200:
        return general_utils.bytes2dict(resp.content)
    
    

def add_category(cat_str: str, color=None):
    cats = cat_str.split('/')
    resp = requests.post(
        cfg.base_url + "/categories",
        json={
            "names": cats,
            "color": color
        },
        headers=auth_utils.build_jwt_header(cfg.config_path)
    )
    return resp
        
        
def remove_category(cat_str: str):
    cats = cat_str.split('/')
    resp = requests.delete(
        cfg.base_url + "/categories",
        json={
            "cats": cats
        },
        headers=auth_utils.build_jwt_header(cfg.config_path)
    )
    
    return resp


def rename_category(cat_str: str, new_name: str):
    cats = cat_str.split('/')
    resp = requests.patch(
        cfg.base_url + "/categories/name",
        json={
            "cats": cats,
            "new_name": new_name
        },
        headers=auth_utils.build_jwt_header(cfg.config_path)
    )
    
    return resp


def move_category(cat_str: str, new_parent_cat_str: str):
    if cat_str.endswith('/'):
        cat_str = cat_str[:-1]
    if new_parent_cat_str.endswith('/'):
        new_parent_cat_str = new_parent_cat_str[:-1]
        
    cats = cat_str.split('/') if cat_str else []
    new_parent_cats = new_parent_cat_str.split('/') if new_parent_cat_str else []
    resp = requests.patch(
        cfg.base_url + "/categories/parent",
        json={
            "cats": cats,
            "new_parent_cats": new_parent_cats
        },
        headers=auth_utils.build_jwt_header(cfg.config_path)
    )
    return resp


def get_last_cat_id(cats: list):
    resp = requests.get(
        cfg.base_url + "/categories/last-cat-id",
        json={
            "cats": cats
        },
        headers=auth_utils.build_jwt_header(cfg.config_path)
    )
    if resp.status_code == 200:
        return general_utils.bytes2dict(resp.content)['cat_id']
    elif resp.status_code == 400:
        err_msg = general_utils.bytes2dict(resp.content)['detail']
        display_utils.center_print(err_msg, constants.DISPLAY_TERMINAL_COLOR_ERROR)
        exit(0)
    else:
        display_utils.center_print(resp.content, constants.DISPLAY_TERMINAL_COLOR_ERROR)
        exit(0)
        
        
def get_category_color(cat_id: int):
    resp = requests.get(
        cfg.base_url + f"/categories/{cat_id}/color",
        headers=auth_utils.build_jwt_header(cfg.config_path)
    )
    
    if resp.status_code == 200:
        return general_utils.bytes2dict(resp.content)['color']
    elif resp.status_code == 400:
        err_msg = general_utils.bytes2dict(resp.content)['detail']
        display_utils.center_print(err_msg, constants.DISPLAY_TERMINAL_COLOR_ERROR)
        exit(0)
    else:
        display_utils.center_print(resp.content, constants.DISPLAY_TERMINAL_COLOR_ERROR)
        exit(0)
    
    
def get_color_dict():
    resp = requests.get(
        cfg.base_url + "/categories/color",
        headers=auth_utils.build_jwt_header(cfg.config_path)
    )
    if resp.status_code == 200:
        color_dict = general_utils.bytes2dict(resp.content)
        return color_dict
    elif resp.status_code == 400:
        err_msg = general_utils.bytes2dict(resp.content)['detail']
        display_utils.center_print(err_msg, constants.DISPLAY_TERMINAL_COLOR_ERROR)
    else:
        display_utils.center_print("Error occurred.", constants.DISPLAY_TERMINAL_COLOR_ERROR)
    


import json
import os
from pathlib import Path

def config_setup(APP_DIR, CONFIG_PATH):
    if not CONFIG_PATH.is_file():
        if not os.path.isdir(APP_DIR):
            os.makedirs(APP_DIR)
        write_json(CONFIG_PATH, {})


def read_json(fpath):
    with open(fpath, 'r') as f:
        data = json.load(f)
        return data
    
    
def write_json(fpath, data):
    with open(fpath, 'w') as f:
        json.dump(data, f)
        

def update_json(fpath, data):
    with open(fpath, 'r') as f:
        org_data = json.load(f)
        org_data.update(**data)
    write_json(fpath, org_data)
    
    
def bytes2dict(b):
    return json.loads(b.decode('utf-8'))


def dict2json(d):
    return json.dumps(d)


def cache_task_ids(cfg, cache: dict):
    target_path = Path(cfg.app_dir) / 'task_id_cache.json'
    write_json(target_path, cache)
    

def read_task_ids_cache(cfg):
    target_path = Path(cfg.app_dir) / 'task_id_cache.json'    
    return read_json(target_path)

    
def exist_task_ids_cache(cfg):
    target_path = Path(cfg.app_dir) / 'task_id_cache.json'
    target_path = os.path.isfile(target_path)

    



from pathlib import Path
import os.path as osp

import typer
from rich import print
from rich.console import Console
from rich.table import Table

from girok.config import get_config
import girok.utils.general as general_utils
import girok.utils.auth as auth_utils
import girok.api.auth as auth_api

app = typer.Typer(rich_markup_mode='rich')

cfg = get_config()

@app.command("login")
def login():
    # Check if the user holds a valid JWT (logged in)
    stored_access_token = auth_utils.get_access_token_from_json(cfg.config_path)
    if stored_access_token:
        resp = auth_api.validate_access_token(stored_access_token)
        if resp.status_code == 200:
            print("You have already logged in. Enjoy!")
            exit(0)
    
    # Log-in
    print("Please login to proceed")
    email = typer.prompt("Enter email address")
    password = typer.prompt("Enter password", hide_input=True)
    
    resp = auth_api.login(email, password)
    if resp.status_code == 200:
        access_token = general_utils.bytes2dict(resp.content)['access_token']
        general_utils.update_json(cfg.config_path, {"access_token": access_token, "email": email})
        print("You're logged in!")
        
    else:
        print("Login failed. Please try again with [green]girok login[/green]")
        exit(0)
    exit(0)
    

@app.command("logout")
def logout():
    access_token = auth_utils.get_access_token_from_json(cfg.config_path)
    if not auth_utils.is_logged_in(access_token):
        print("You're not logged in.")
        exit(0)
    
    auth_utils.remove_access_token(cfg.config_path)
    print("Successfully logged out")
    exit(0)
    
    
    
@app.command("register")
def register():
    access_token = auth_utils.get_access_token_from_json(cfg.config_path)
    if access_token and auth_utils.is_logged_in(access_token):
        print("You're logged in. Please try after log out.")
        exit(0)
        
    print(":star: This is your first time to use [yellow]girok[/yellow]! :star:\n")
    is_register = typer.confirm("Do you want to register a new account?")
    if is_register:
        email = typer.prompt("Enter email address")
        password = typer.prompt("Enter password", hide_input=True)
        confirm_password = typer.prompt("Confirm password", hide_input=True)
        auth_utils.match_passwords(password, confirm_password)
        
        # register a new account
        register_resp = auth_api.register(email, password)
        if register_resp.status_code == 201:
            print("Register complete!")
        else:
            register_resp_content = general_utils.bytes2dict(register_resp.content)
            if register_resp.status_code == 400:
                print(register_resp_content['detail'])
            exit(0)
        
        print("Registration successful! Please login by [green]girok login[/green] command")
        exit(0)
    else:
        exit(0)

if __name__ == '__main__':
    app()


from datetime import datetime, timedelta
from calendar import monthrange
import calendar

def is_valid_year(year: int):
    if year < datetime.now().year - 3 or year > datetime.now().year + 10:
        return False
    return True
    
    
def is_valid_month(month: int):
    if month <= 0 or month > 12:
        return False
    return True


def is_valid_day(year: int, month: int, day: int):
    if day not in range(1, monthrange(year, month)[1] + 1):
        return False
    return True


def is_valid_hour(hour: int):
    if hour < 0 or hour > 23:
        return False
    return True


def is_valid_minute(minute: int):
    if minute < 0 or minute > 59:
        return False
    return True


def get_date_from_shortcut(this_week: bool, weekday_num: int):
    """
    Params:
        - this_week (bool): True if this week, False if next week
        - weekday_num (int): 0 ~ 6, referring to monday ~ sunday
    """
    if not this_week:
        weekday_num += 7
        
    current_weekday_num = datetime.now().weekday()
    day_offset = weekday_num - current_weekday_num
    deadline_date = datetime.now() + timedelta(days=day_offset)
    
    return deadline_date.year, deadline_date.month, deadline_date.day


def get_date_from_offset(offset: int):
    deadline_date = datetime.now() + timedelta(days=offset)
    return deadline_date.year, deadline_date.month, deadline_date.day


def get_today_date():
    now = datetime.now()
    return now.year, now.month, now.day

def get_last_day_of_current_month():
    now = datetime.now()
    return monthrange(now.year, now.month)[1]


def build_time_window(start: list, end: list):
    """
    start (list): [year, month, day]
    end (list): [year, month, day]
    """
    start_date, end_date = f"{start[0]}-{start[1]}-{start[2]} 00:00:00", f"{end[0]}-{end[1]}-{end[2]} 11:59:59" 
    return start_date, end_date


def build_time_window_by_day_offsets(start_offset: int, end_offset: int):
    start = get_date_from_offset(start_offset)
    end = get_date_from_offset(end_offset)
    return build_time_window([*start], [*end])


def build_time_window_by_month_offset(month_offset: int):
    now = datetime.now()
    current_year, current_month = now.year, now.month
    new_year, new_month = get_year_and_month_by_month_offset(month_offset)
    start_date = f"{current_year}-{current_month}-01 00:00:00"
    last_day_of_end_month = monthrange(new_year, new_month)[1]
    end_date = f"{new_year}-{new_month}-{last_day_of_end_month} 11:59:59"
    return start_date, end_date


def build_current_month_time_window():
    now = datetime.now()
    year, month = now.year, now.month
    start_date = f"{year}-{month}-01 00:00:00"
    last_day_of_current_month = get_last_day_of_current_month()
    end_date = f"{year}-{month}-{last_day_of_current_month} 11:59:59"
    return start_date, end_date


def build_time_window_by_year_and_month(year=None, month=None):
    start_date = f"{year}-{month}-01 00:00:00"
    last_day_of_current_month = monthrange(year, month)[1]
    end_date = f"{year}-{month}-{last_day_of_current_month} 11:59:59"
    return start_date, end_date
    

def get_year_and_month_by_month_offset(month_offset: int, year=None, month=None):
    if year is None and month is None:
        now = datetime.now()
        year, month = now.year, now.month
    new_months = (year * 12 + month + month_offset)
    new_year = new_months // 12
    new_month = new_months % 12
    if new_month == 0: # (1 - 1) % 12 = 0.
        new_year -= 1
        new_month = 12
    return new_year, new_month


def get_current_weekday_num():
    now = datetime.now()
    current_weekday_num = now.weekday()
    return current_weekday_num


def get_weekday_name_from_date(year, month, day, abbr=True):
    time = datetime.strptime(f"{year}-{month}-{day}", "%Y-%m-%d")
    if abbr:
        return calendar.day_abbr[time.weekday()]
    else:
        return calendar.day_name[time.weekday()]


def get_month_name_from_month(month: int, abbr=True):
    if abbr:
        return get_month_name_by_number(month, abbr=True)
    else:
        return get_month_name_by_number(month, abbr=False)


def build_date_info(dt: datetime):
    year, month, day = dt.year, dt.month, dt.day
    h, m, s = str(dt.time()).split(":")
    month_name = get_month_name_from_month(month, abbr=True)
    weekday_name = get_weekday_name_from_date(year, month, day, abbr=False)
    return f"{weekday_name}, {month_name} {day}, {year}"


def get_day_offset_between_two_dates(dt1, dt2):
    dt1 = datetime.strptime(f"{dt1.year}-{dt1.month}-{dt1.day}", "%Y-%m-%d")
    dt2 = datetime.strptime(f"{dt2.year}-{dt2.month}-{dt2.day}", "%Y-%m-%d")
    diff = dt2 - dt1
    return diff.days

def get_month_name_by_number(month_num: int, abbr=False):
    """
    month_num: 0 ~ 11
    """
    if abbr:
        return calendar.month_abbr[month_num]
    else:
        return calendar.month_name[month_num]


from textual.app import App, ComposeResult
from textual.containers import Container, Horizontal, Vertical
from textual.widgets import Button, Footer, Header, Static, Label, Placeholder, Tree
from textual.messages import Message
from textual.widget import Widget
from textual.reactive import var, reactive
from textual import events

import girok.api.category as category_api
import girok.api.task as task_api
import girok.utils.calendar as calendar_utils
import girok.utils.general as general_utils
import girok.constants as constants
from rich.style import Style
from rich.text import Text
from textual import log
from textual.widgets._tree import TreeNode

class CategoryTree(Tree):
    CSS_PATH = "./demo_dock.css"
    cats = dict()
    can_focus = True
    can_focus_children = True
    auto_expand=False
    highlighted_node = None
    selected_node = None
    init_select = False
    init_highlight = False
    class CategoryChanged(Message):
        def __init__(self, cat_path: str):
            super().__init__()
            self.cat_path = cat_path

    class CustomTestMessage(Message):
        def __init__(self):
            super().__init__()
    
    def on_mount(self):
        # Select the topmost node
        # self.select_node(self.root)
        self.highlighted_node = self.root
        self.selected_node = self.root
        # calendar_utils.add_highlight(self.selected_node)
        calendar_utils.add_left_arrow_tree(self.highlighted_node)
        # self.temp_node_selected()
        # self.select_node(self.selected_node)
        # self.action_select_cursor()

        self.line = 0
        self.cats = category_api.get_categories()
        self.root.expand()
        
        for cat in self.cats:
            top_cat = self.root.add(cat, expand=True, data={"color": self.cats[cat]['color']})
            top_cat.allow_expand = True
            calendar_utils.build_category_tree(top_cat, self.cats[cat]['subcategories'])
            
    def on_key(self, evt):
        if evt.key == "j":
            self.action_cursor_down()
        elif evt.key == "k":
            self.action_cursor_up()
        elif evt.key == "o":
            self.action_select_cursor()

    def render_label(self, node, base_style: Style, style: Style):
        node_label = node._label.copy()
        icon = ""
        if node.parent is None:
            icon = "📖 "
        elif node.parent.parent is None:
            icon = Text("● ", style=constants.CIRCLE_COLOR[node.data['color']])
        text = Text()
        text.append(icon)
        text.append(node_label)
        return text
        
    def on_tree_node_selected(self, event: Tree.NodeSelected):
        event.stop()
        # calendar_utils.remove_highlight(self.selected_node)
        # calendar_utils.add_highlight(event.node)
        full_cat_path = calendar_utils.get_full_path_from_node(event.node)
        self.selected_node = event.node
        self.post_message(self.CategoryChanged(full_cat_path))
        
        
    def on_tree_node_highlighted(self, event: Tree.NodeHighlighted):
        event.stop()
        prev_highlighted_node = self.highlighted_node
        calendar_utils.remove_left_arrow_tree(prev_highlighted_node)
        calendar_utils.remove_highlight(prev_highlighted_node)
        calendar_utils.add_left_arrow_tree(event.node)
        calendar_utils.add_highlight(event.node)
        self.highlighted_node = event.node
        
    def on_focus(self, evt):
        calendar_utils.add_left_arrow_tree(self.highlighted_node)
        
        
class TagTree(Tree):
    CSS_PATH = "./demo_dock.css"
    tags = []
    can_focus = True
    can_focus_children = True
    auto_expand=False
    highlighted_node = None
    selected_node = None
    
    class TagChanged(Message):
        def __init__(self, tag: str):
            super().__init__()
            self.tag = tag
            
    class CustomTestMessage(Message):
        def __init__(self):
            super().__init__()
            
    def on_mount(self):
        self.select_node(self.root)
        self.action_select_cursor()
        self.highlighted_node = self.root
        self.selected_node = self.root
        
        resp = task_api.get_tags()
        if resp.status_code == 200:
            self.tags = general_utils.bytes2dict(resp.content)['tags']
        elif resp.status_code == 400:
            err_msg = general_utils.bytes2dict(resp.content)['detail']
            exit(0)
        else:
            exit(0)
        
        self.root.expand()
        
        for tag in self.tags:
            self.root.add(tag, expand=True)
            
    def on_key(self, evt):
        if evt.key == "j":
            self.action_cursor_down()
        elif evt.key == "k":
            self.action_cursor_up()
        elif evt.key == "o":
            self.action_select_cursor()
            
    def on_focus(self, evt):
        calendar_utils.add_left_arrow_tree(self.highlighted_node)
            
    def render_label(self, node, base_style: Style, style: Style):
        node_label = node._label.copy()
        
        icon = ""
        if node.parent is None:
            icon = "📖 "
        elif node.parent.parent is None:
            icon = Text("● ", style="white")
                
        text = Text()
        text.append(icon)
        text.append(node_label)
        return text
        
    def on_tree_node_selected(self, event: Tree.NodeSelected):
        event.stop()
        # calendar_utils.remove_highlight(self.selected_node)
        tag = str(event.node._label)
        if tag.endswith(" " + constants.LEFT_ARROW_EMOJI):
            tag = tag[:-2]

        self.post_message(self.TagChanged(tag))
        event.node.set_label(tag)
        # calendar_utils.add_highlight(event.node)
        self.selected_node = event.node
        # self.post_message(self.CustomTestMessage())
        
    def on_tree_node_highlighted(self, event: Tree.NodeHighlighted):
        event.stop()
        prev_highlighted_node = self.highlighted_node
        calendar_utils.remove_highlight(prev_highlighted_node)
        calendar_utils.remove_left_arrow_tree(prev_highlighted_node)
        calendar_utils.add_left_arrow_tree(event.node)
        calendar_utils.add_highlight(event.node)
        self.highlighted_node = event.node
        
        
class SidebarMainContainer(Vertical):
    CSS_PATH = "./demo_dock.css"
    def compose(self):
        yield CategoryTree("All Categories", id="sidebar")
        yield TagTree("All Tags", id="tag-tree")
    
    
class SidebarContainer(Vertical):
    CSS_PATH = "./demo_dock.css"
        
    def compose(self):
        yield SidebarMainContainer(id="sidebar-main-container")

        

        



import asyncio
from datetime import datetime, timedelta
import calendar

from textual.app import App, ComposeResult
from textual.containers import Container, Horizontal, Vertical
from textual.widgets import Button, Footer, Header, Static, Label, Placeholder, Tree, DataTable
from textual.messages import Message
from textual.widget import Widget
from textual.reactive import var
from rich.text import Text
from rich.style import Style
from rich.segment import Segment
from rich.panel import Panel
from rich.markdown import Markdown
from textual import log

import girok.api.task as task_api
import girok.api.category as category_api
import girok.utils.calendar as calendar_utils
import girok.utils.general as general_utils
import girok.utils.task as task_utils
import girok.constants as constants
from girok.calendar_cli.sidebar import CategoryTree, TagTree


class WeekdayBarContainer(Horizontal):
    pass

class CalendarHeader(Vertical):
    year = datetime.now().year
    month = datetime.now().month
    cat_path = ""
    tag = ""

    def on_mount(self):
        self.display_date()

    def compose(self):
        month_name = task_utils.get_month_name_by_number(self.month)

        with Horizontal():
            with Container(id="calendar-header-category-container"):
                yield Static(self.cat_path, id="calendar-header-category")
            with Container(id="calendar-header-date-container"):
                yield Static(Text(f"{month_name} {self.year}", style=Style(color=constants.CALENDAR_HEADER_DATE_COLOR, bold=True)), id="calendar-header-date")
            with Container(id="calendar-header-tag-container"):
                yield Static(f"{self.tag}", id="calendar-header-tag")
        yield Horizontal()
        with WeekdayBarContainer(id="weekday-bar"):
            yield Static(Text("Monday", style=Style(color=constants.CALENDAR_WEEKDAY_NAME_COLOR, bold=True)), classes="calendar-weekday-name")
            yield Static(Text("Tuesday", style=Style(color=constants.CALENDAR_WEEKDAY_NAME_COLOR, bold=True)), classes="calendar-weekday-name")
            yield Static(Text("Wednesday", style=Style(color=constants.CALENDAR_WEEKDAY_NAME_COLOR, bold=True)), classes="calendar-weekday-name")
            yield Static(Text("Thursday", style=Style(color=constants.CALENDAR_WEEKDAY_NAME_COLOR, bold=True)), classes="calendar-weekday-name")
            yield Static(Text("Friday", style=Style(color=constants.CALENDAR_WEEKDAY_NAME_COLOR, bold=True)), classes="calendar-weekday-name")
            yield Static(Text("Saturday", style=Style(color="#87C5FA", bold=True)), classes="calendar-weekday-name")
            yield Static(Text("Sunday", style=Style(color="#DB4455", bold=True)), classes="calendar-weekday-name")

    def update_year_and_month(self, year, month):
        self.year, self.month = year, month
        self.display_date()

    def update_cat_path(self, new_cat_path):
        self.cat_path = new_cat_path
        self.display_date()

    def update_tag(self, new_tag):
        self.tag = new_tag
        self.display_date()

    def display_date(self):
        month_name = task_utils.get_month_name_by_number(self.month, abbr=False)
        calendar_header_category = self.query_one("#calendar-header-category")
        calendar_header_date = self.query_one("#calendar-header-date")
        calendar_header_tag = self.query_one("#calendar-header-tag")
        calendar_weekday_bar = self.query_one("#weekday-bar")

        calendar_header_category.update(Text(f"Category: /{self.cat_path}", style=Style(color=constants.CALENDAR_HEADER_DATE_COLOR)))
        calendar_header_date.update(Text(f"{month_name} {self.year}", style=Style(color=constants.CALENDAR_HEADER_DATE_COLOR, bold=True)))
        calendar_header_tag.update(Text(f"Tag: {self.tag}", style=Style(color=constants.CALENDAR_HEADER_DATE_COLOR)))


class CalendarCell(Vertical):
    pass


class Calendar(Container):
    year = datetime.now().year
    month = datetime.now().month
    cat_path = "" # If "", show all categories
    tag = ""
    tasks = []
    can_focus = True
    cur_month_first_day_cell_num = None
    cur_focused_cell_cord = (None, None)
    cur_focused_cell = None
    m = 5
    n = 7
    grid = [[False for _ in range(7)] for _ in range(5)]
    is_pop_up = False

    class TaskCellSelected(Message):
        def __init__(self, cell_tasks, year, month, day):
            super().__init__()
            self.cell_tasks = cell_tasks
            self.year = year
            self.month = month
            self.day = day

    def on_mount(self):
        self.update_calendar()

    def compose(self):
        for i in range(35):
            yield CalendarCell(classes="calendar-cell", id=f"cell{i}")

    def on_key(self, event):
        if self.is_pop_up:
            return
        x, y = self.cur_focused_cell_cord
        if event.key == 'h': # left
            next_cell_coord = (x, y - 1)
        elif event.key == 'j': # down
            next_cell_coord = (x + 1, y)
        elif event.key == 'k': # up
            next_cell_coord = (x - 1, y)
        elif event.key == 'l': # right
            next_cell_coord = (x, y + 1)
        elif event.key == 'o':
            pass
        else:
            return

        if event.key in ['h', 'j', 'k', 'l']: # moving on cells
            nx, ny = next_cell_coord
            if nx < 0 or ny < 0 or nx >= self.m or ny >= self.n: # Out of matrix
                return

            if not self.grid[nx][ny]: # Out of boundary of the current month
                return

            prev_cell_num = calendar_utils.convert_coord_to_cell_num(*self.cur_focused_cell_cord)
            prev_cell = self.query_one(f"#cell{prev_cell_num}")
            calendar_utils.remove_left_arrow(prev_cell)

            cur_cell_num = calendar_utils.convert_coord_to_cell_num(nx, ny)
            next_cell = self.query_one(f"#cell{cur_cell_num}")
            calendar_utils.add_left_arrow(next_cell)

            self.cur_focused_cell_cord = (nx, ny)
            self.cur_focused_cell = next_cell
        elif event.key == 'o': # select a cell
            cur_cell_num = calendar_utils.convert_coord_to_cell_num(x, y)
            cur_cell = self.query_one(f"#cell{cur_cell_num}")
            cell_tasks = cur_cell.children[1:] # task data

            # Retrieve tasks for the selected day
            selected_day = calendar_utils.convert_cell_num_to_day(self.year, self.month, cur_cell_num)
            cell_tasks = list(filter(
                lambda x: calendar_utils.get_date_obj_from_str_separated_by_T(x['deadline']).day == selected_day,
                self.tasks
            ))
            self.post_message(self.TaskCellSelected(cell_tasks, self.year, self.month, selected_day))
            self.is_pop_up = True

    def on_focus(self):
        x, y = calendar_utils.convert_cell_num_to_coord(self.cur_month_first_day_cell_num)
        target_cell = self.query_one(f"#cell{self.cur_month_first_day_cell_num}")
        # target_cell.add_class("focused-cell")
        calendar_utils.add_left_arrow(target_cell)
        self.cur_focused_cell_cord = (x, y)
        self.cur_focused_cell = target_cell

    def update_year_and_month(self, year, month):
        self.year, self.month = year, month
        self.update_calendar()

    def update_cat_path(self, new_cat_path: str):
        self.cat_path = new_cat_path
        self.update_calendar(show_arrow=False)

    def update_tag(self, new_tag: str):
        self.tag = new_tag
        self.update_calendar(show_arrow=False)

    def refresh_cell_days(self):
        self.grid = [[False for _ in range(7)] for _ in range(5)]
        first_weekday, total_days = calendar.monthrange(self.year, self.month)
        self.cur_month_first_day_cell_num = first_weekday
        now = datetime.now()
        for i in range(35):
            cell = self.query_one(f"#cell{i}")
            for child in cell.walk_children():
                child.remove()
            if i >= first_weekday and i <= first_weekday + total_days - 1:
                x, y = calendar_utils.convert_cell_num_to_coord(i)
                self.grid[x][y] = True
                day = calendar_utils.convert_cell_num_to_day(self.year, self.month, i)
                day_text = Text()
                day_text.append(f"{day}")
                if self.year == now.year and self.month == now.month and day == now.day:
                    day_text = Text(str(day_text), style=Style(bgcolor=constants.CALENDAR_TODAY_COLOR, color="black"))
                    # day_text.stylize(style=Style(bgcolor=constants.CALENDAR_TODAY_COLOR, color="black"))
                cell.mount(Label(day_text, id=f"cell-header-{i}"))

    def update_calendar(self, show_arrow=True):
        """
        If val == "", then "root category" is selected
        """
        if self.cat_path == "": # all categories
            cat_list = None
        elif self.cat_path == "No Category":
            cat_list = ['']
        else:
            cat_list = self.cat_path[:-1].split('/')

        if self.tag == "":
            tag = None
        else:
            tag = self.tag
            
        log("CATS", cat_list),
        log("tag", tag)

        start_date, end_date = task_utils.build_time_window_by_year_and_month(self.year, self.month)
        resp = task_api.get_tasks(
            cats=cat_list,
            start_date=start_date,
            end_date=end_date,
            tag=tag,
            view="list"
        )

        if resp.status_code == 200:
            self.refresh_cell_days()
            tasks = general_utils.bytes2dict(resp.content)['tasks']
            self.tasks = tasks
            first_weekday, last_day = calendar.monthrange(self.year, self.month)

            if self.cur_focused_cell:
                if show_arrow:
                    calendar_utils.remove_left_arrow(self.cur_focused_cell)
                self.cur_focused_cell_cord = calendar_utils.convert_cell_num_to_coord(first_weekday)
                self.cur_focused_cell = self.query_one(f"#cell{first_weekday}") # update
                if show_arrow:
                    calendar_utils.add_left_arrow(self.cur_focused_cell)

            for idx, task in enumerate(tasks):
                full_date = task['deadline']
                day = datetime.strptime(full_date, "%Y-%m-%dT%H:%M:%S").day
                cell_num = calendar_utils.convert_day_to_cell_num(self.year, self.month, day)
                cell = self.query_one(f"#cell{cell_num}")
                color = task['color']
                name = task['name']
                # if len(name) > 13:
                #     name = name[:13] + ".."
                task_item_name = Text()
                task_item_name.append("●", style=constants.CIRCLE_COLOR[color])
                task_item_name.append(" " + name)

                task_item = Static(task_item_name, id=f"task-cell{cell_num}-{idx}", classes="task-item")
                cell.mount(task_item)

                task_item = self.query_one(f"#task-cell{cell_num}-{idx}")
                task_item.styles.overflow_x = "hidden"
                task_item.styles.overflow_y = "hidden"

        elif resp.status_code == 400:
            err_msg = general_utils.bytes2dict(resp.content)['detail']
            exit(0)
        else:
            exit(0)


class CalendarContainer(Vertical):
    year = datetime.now().year
    month = datetime.now().month
    cat_path = None # If none, show all categories
    tag = None

    def update_month_by_offset(self, offset: int):
        new_year, new_month = task_utils.get_year_and_month_by_month_offset(
            month_offset=offset,
            year=self.year,
            month=self.month
        )
        self.year, self.month = new_year, new_month
        calendar_header = self.query_one(CalendarHeader)
        cal = self.query_one(Calendar)

        calendar_header.update_year_and_month(self.year, self.month)
        cal.update_year_and_month(self.year, self.month)

    def update_year_and_month(self, year: int, month: int):
        self.year, self.month = year, month
        calendar_header = self.query_one(CalendarHeader)
        cal = self.query_one(Calendar)

        calendar_header.update_year_and_month(self.year, self.month)
        cal.update_year_and_month(self.year, self.month)

    def update_cat_path(self, cat_path: str):
        """
        cat_path: ex) HKU/COMP3230 or ""
        """
        self.cat_path = cat_path
        cal = self.query_one(Calendar)
        cal.update_cat_path(new_cat_path=cat_path)

        cal_header = self.query_one(CalendarHeader)
        cal_header.update_cat_path(new_cat_path=cat_path)


    def update_tag(self, tag: str):
        self.tag = tag
        cal = self.query_one(Calendar)
        cal.update_tag(new_tag=tag)
        cal_header = self.query_one(CalendarHeader)
        cal_header.update_tag(new_tag=tag)


    def compose(self):
        yield CalendarHeader(id="calendar-header")
        yield Calendar(id="calendar")
